//
//  XSPEC12  November 2003
//
//

#include <xsTypes.h>
#include <XSContainer.h>
#include <XSFit/Fit/Fit.h>
#include <XSModel/Parameter/Parameter.h>  
#include <XSModel/Parameter/ResponseParam.h>
#include <XSUser/Global/Global.h>
#include <XSUser/Global/XSGlobal.h>
#include <XSUser/Handler/HandlerUtils.h>
#include <XSUser/UserInterface/xstcl.h>
#include <XSUtil/Parse/XSparse.h>
#include <sstream>

namespace
{
   // Strategy class needed for freezeThawUntie template function.
   class ThawStrategy
   {
      public:
         static bool perform(Parameter* par, string& msg);
   };
}

int
XSGlobal::xsThaw(ClientData cdata,Tcl_Interp* tclInterp,int objc, Tcl_Obj* CONST  objv[])
{
   using namespace XSContainer;

   // prevModelName is only ever applied when user enters command with no args,
   // in which case the previous <modName>:<range> specifier (if any) is used.
   // prevRange can also be applied when using single '*' notation in range.
   RangePair prevRange = HandlerUtils::FreezeThawRange::getRange();
   const string& prevModelName = HandlerUtils::FreezeThawRange::getGroupName();
   
   bool isAnyChanged = false;
   const bool isRespPar = (string(Tcl_GetString(objv[0])) == string("rthaw"));
   try
   {
      string defName;
      StringArray inArgs;
      if (objc == 1)
      {
         if (prevRange.second == 0)
         {
            printDocs(static_cast<char*>(cdata),"?");
            return globalData->autoSave(TCL_OK);
         }
         else
         {
            defName = prevModelName;
            std::ostringstream oss;
            oss << prevRange.first;
            if (prevRange.second > prevRange.first)
               oss << "-" << prevRange.second;
            inArgs.push_back(oss.str());
         }
      }   
      else
      {
         string arg(Tcl_GetString(objv[1]));
         if (arg[0] == '?') 
         {
            printDocs(static_cast<char*>(cdata),"?");
            return globalData->autoSave(TCL_OK);
         }
         else
         {
            for (int i=1; i<objc; ++i)
               inArgs.push_back(string(Tcl_GetString(objv[i])));
         }      
      }
      
      
      // Break up input args into groups of StringArrays, where
      // each group belongs to a particular model.  Then we
      // can use the getRanges function to do the grunt work.
      std::map<string, StringArray> modelArgGroups;
      XSparse::collateRanges(inArgs, defName, modelArgGroups);
      
      std::map<string, StringArray>::iterator itModGroups = 
             modelArgGroups.begin();
      std::map<string, StringArray>::iterator itModGroupsEnd =
             modelArgGroups.end();
      while (itModGroups != itModGroupsEnd)
      {
         bool isChanged = false;
         if (isRespPar)
            isChanged = freezeThawUntie<ResponseParam,ThawStrategy>
                              (itModGroups->first,itModGroups->second, prevRange);
         else
            isChanged = freezeThawUntie<Parameter,ThawStrategy>
                              (itModGroups->first,itModGroups->second, prevRange);
         if (isChanged)
            isAnyChanged = true;
         
         HandlerUtils::FreezeThawRange::setRange(prevRange);
         
         // For xsFreeze, xsThaw, and xsUntie: Deliberately incrementing itModGroups 
         // this way rather than on a separate line.  This gets around a strange -O2 
         // level compiler bug that first appeared only with a build using g++ 4.2.1 
         // on powerpc-apple-darwin9.5.0, and seems to be triggered by the above 
         // template instantiations being enclosed in a loop.  
         HandlerUtils::FreezeThawRange::setGroupName((itModGroups++)->first);
      }
                     
      if (isAnyChanged)
      {
         XSContainer::fit->Update();
      }
      return globalData->autoSave(TCL_OK);     
   }
   catch (YellowAlert&)
   {
      if (isAnyChanged)
      {
         XSContainer::fit->Update();
      }
      return globalData->autoSave(TCL_ERROR);       
   }      
        
}


namespace
{
   bool ThawStrategy::perform(Parameter* par, string& msg)
   {
      bool isPerformed = false;
      if (!par->isFrozen())
      {
         msg = string(" is not frozen.");
      }
      else
      {
         // Unlike freeze, don't enclose in try/catch.  If something
         // throws below, it is a more serious problem and should
         // therefore break out of loops.  If parameter is not a
         // ModParam this simply does nothing.
         par->thaw();
         isPerformed = true;
      }
      return isPerformed;
   }
}
